/* Copyright (C) 2005-2008 Damien Stehle.
Copyright (C) 2007 David Cade.

This file is part of the fplll Library.

The fplll Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The DPE Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the DPE Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. */


/* prec=53, eta=0.501, dim < dim_double_max [ (delta / 100.0) + 25 ] */
static const double dim_double_max[75]=
  {0,26,29.6,28.1,31.1,32.6,34.6,34,37.7,38.8,39.6,41.8,40.9,43.6,44.2,47,46.8,50.6,49.1,51.5,52.5,54.8,54.6,57.4,57.6,59.9,61.8,62.3,64.5,67.1,68.8,68.3,69.9,73.1,74,76.1,76.8,80.9,81.8,83,85.3,87.9,89,90.1,89,94.6,94.8,98.7,99,101.6,104.9,106.8,108.2,107.4,110,112.7,114.6,118.1,119.7,121.8,122.9,126.6,128.6,129,133.6,126.9,135.9,139.5,135.2,137.2,139.3,142.8,142.4,142.5,145.4};

static const double eta_dep[10]=
  {1.,//0.5
   1.,//0.55
   1.0521,//0.6
   1.1254,//0.65
   1.2535,//0.7
   1.3957,//0.75
   1.6231,//0.8
   1.8189,//0.85
   2.1025,//0.9
   2.5117};//0.95



inline void present(const char* s, int p)
{
#ifdef VERBOSE
  cerr<<"\n====== LLL method : "<<s;
  if (p)
    cerr<<" (precision = "<<p<<")";
  cerr<<" ======\n\n";
#endif
}
inline void echecpresent(int kappa)
{
#ifdef VERBOSE
  cerr<<"\n====== LLL method end : ";
  if (kappa==0)
    cerr<<"success";
  else if (kappa>0)
    cerr<<"Size-reduction failed. (kappa="<<kappa<<")";
  else
    cerr<<"Too many iterations.";

  cerr<<" ======\n\n";
#endif
}
inline int min(int x,int y)
{
  return x<y?x:y;
}

wrapper::wrapper(ZZ_mat<mpz_t>*B,int precision,double eta_,double delta_)
{

  B1=B;
  eta=eta_;
  delta=delta_;

  max_exp = B->getMaxExp();
  n = B->GetNumCols();
  d = B->GetNumRows();

  double rho = (1.0 + eta) * (1.0+eta) / (delta - eta*eta);
  
  unsigned int goodprec = (unsigned int) (7.0 + 0.2*d + d* log(rho)/ log(2.0) 
			     +  2.0 * log ( (double) d )  
			     - log( (eta-0.5)*(1.0-delta) ) / log(2.));
  prec=(precision==0 ? goodprec : precision);

  max_exp_current = max_exp + ((int) (0.5*log((double) d*n)/log(2.)))+1;


  if (max_exp_current < CPU_SIZE-2)
    B2 = new ZZ_mat<long int>(d, n);

}

bool wrapper::little(int kappa,int precision)
{
  /*one may add here dimension arguments with respect to eta and delta */
  int dm=(int)(delta*100.-25.);
  if (dm<0) dm=0;
  if (dm>74) dm=74;

  int em=(int) ((eta-0.5)*20);
  if (em<0) em=0;
  if (em>9) em=9;
  
  double p=precision<53?53.:((double)precision);
  p/=53.0;
  
  p*=eta_dep[em] ;/*eta dependance*/
  p*=dim_double_max[dm];
#if 0
#ifdef VERBOSE
  cerr << kappa << " compared to " << p << "\n";
#endif
#endif
  return kappa<p;
}



int wrapper::heuristicLoop(int precision)
{
  int kappa;
  if (precision<=PREC_DOUBLE)
    {
      if (max_exp_current>MAX_EXP_DOUBLE)
	{
	  present("heuristic<mpz_t, dpe_t>",0);
	  kappa=heuristic<mpz_t,dpe_t>(B1,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else if (max_exp_current>CPU_SIZE -2)
	{
	  present("heuristic<mpz_t, double>",0);
	  kappa=heuristic<mpz_t,double>(B1,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else
	{
	  present("heuristic<long int, double>",0);
	  kappa=heuristic<long int,double>(B2,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
    }
  else
    {
      if (max_exp_current > CPU_SIZE-2)
	{
	  present("heuristic<mpz_t, mpfr_t>",precision);
	  kappa=heuristic<mpz_t,mpfr_t>(B1,precision,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else
	{
	  present("heuristic<long int, mpfr_t>",precision);
	  kappa=heuristic<long int,mpfr_t>(B2,precision,eta,delta).LLL();
	  echecpresent(kappa);
	}
    }

  if (little(kappa,precision) && kappa!=0)
    return provedLoop(precision);
  else if (kappa!=0 && precision<prec)
    return heuristicLoop(min(precision*2,prec));
  // this point should never be reached
  return -1;
}
int wrapper::provedLoop(int precision)
{
  int kappa;
  if (precision<=PREC_DOUBLE)
    {
      if (2*max_exp_current>MAX_EXP_DOUBLE)
	{
	  present("proved<mpz_t, dpe_t>",0);
	  kappa=proved<mpz_t,dpe_t>(B1,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else if (2*max_exp_current>CPU_SIZE -2)
	{
	  present("proved<mpz_t, double>",0);
	  kappa=proved<mpz_t,double>(B1,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else
	{
	  present("proved<long int, double>",0);
	  kappa=proved<long int,double>(B2,0,eta,delta).LLL();
	  echecpresent(kappa);
	}
    }
  else
    { 
      if (2*max_exp_current > CPU_SIZE-2)
	{
	  present("proved<mpz_t, mpfr_t>",precision);
	  kappa=proved<mpz_t,mpfr_t>(B1,precision,eta,delta).LLL();
	  echecpresent(kappa);
	}
      else
	{
	  present("proved<long int, mpfr_t>",precision);
	  kappa=proved<long int,mpfr_t>(B2,precision,eta,delta).LLL();
	  echecpresent(kappa);
	}
    }
  
  if (little(kappa,precision) && kappa!=0)
    return kappa;
  else if (kappa!=0 && precision*2<prec)
    return provedLoop(precision*2);
  // this point should never be reached
  return -1;
}

int wrapper::LastLLL()
{
  int kappa, i, j;
  signed long int ltmp;
  Z_NR<mpz_t> ztmp2;

  if (2*max_exp_current < CPU_SIZE -2){
    if (prec<=PREC_DOUBLE)
      {
	present("proved<long int, double>(prec<=PREC_DOUBLE)",prec);
	kappa=proved<long int,double>(B2,prec,eta,delta).LLL();
	echecpresent(kappa);
      }
    else
      {
	present("proved<long int, mpfr_t>",prec);
	kappa=proved<long int,mpfr_t>(B2,prec,eta,delta).LLL();
	echecpresent(kappa);
      }
    for (i=0; i<d; i++)
      for (j=0; j<n; j++)
	{
	  ltmp = (B2->Get(i,j)).GetData();
	  ztmp2.set_si(ltmp);
	  B1->Set(i, j, ztmp2);
	}
    B2->clear();
  }

  else{
    if (prec<=PREC_DOUBLE)
      {
	present("proved<mpz_t, dpe_t>(prec<=PREC_DOUBLE)",prec);
	kappa=proved<mpz_t,dpe_t>(B1,prec,eta,delta).LLL();
	echecpresent(kappa);
      }
    else
      {
	present("proved<mpz_t, mpfr_t>",prec);
	kappa=proved<mpz_t,mpfr_t>(B1,prec,eta,delta).LLL();
	echecpresent(kappa);
      }

  }

  return kappa;
}

int wrapper::LLL()
{
  int kappa, i, j;
  Z_NR<mpz_t> ztmp2;
  Z_NR<long int> ltmp2;
  signed long int ltmp;


  if (max_exp_current<CPU_SIZE - 2)
    {
      for (i=0; i<d; i++)
	for (j=0; j<n; j++)
	  {
	    ltmp = mpz_get_si((B1->Get(i,j)).GetData());
	    ltmp2.set_si(ltmp);
	    B2->Set(i, j, ltmp2);
	  }

      present("fast<long int, double>",0);
      kappa=fast<long int,double>(B2,0,eta,delta).LLL();
      echecpresent(kappa);

      if (little(kappa,53)&& kappa!=0){
	if (2*max_exp_current >= CPU_SIZE - 2){
	  for (i=0; i<d; i++)
	    for (j=0; j<n; j++)
	      {
		ltmp = (B2->Get(i,j)).GetData();
		ztmp2.set_si(ltmp);
		B1->Set(i, j, ztmp2);
	      }
	  B2->clear();
	}
	
	kappa=provedLoop(PREC_DOUBLE);
      }

      else if (kappa!=0)
	{
	  kappa=heuristicLoop(min(PREC_DOUBLE*2,prec));
	  if (2*max_exp_current >= CPU_SIZE - 2){
	    for (i=0; i<d; i++)
	      for (j=0; j<n; j++)
		{
		  ltmp = (B2->Get(i,j)).GetData();
		  ztmp2.set_si(ltmp);
		  B1->Set(i, j, ztmp2);
		}
	    B2->clear();
	  }
	}
    }
  else
    {
      present("fast<mpz_t, double>",0);
      kappa=fast<mpz_t,double>(B1,0,eta,delta).LLL();
      echecpresent(kappa);
      if (little(kappa,53) && kappa!=0)
	kappa=provedLoop(PREC_DOUBLE);
      else if (kappa!=0)
	kappa=heuristicLoop(min(PREC_DOUBLE*2,prec));
    }

  return LastLLL();
}
